#
# Copyright 2020 Osia authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module implements configuration object for aws installation"""
import logging

import boto3
from botocore.exceptions import ClientError

from .base import AbstractInstaller


def _get_connection(*args, **kwargs):
    return boto3.client("ec2", *args, **kwargs)


class AWSInstaller(AbstractInstaller):
    """Object containing all configuration related
    to aws installation"""

    def __init__(self, cluster_region=None, list_of_regions=None,
                 aws_access_key_id=None, aws_secret_access_key=None, **kwargs):
        super().__init__(**kwargs)
        self.cluster_region = cluster_region
        self.list_of_regions = list_of_regions if list_of_regions else []

        self.boto_kwargs = {
            "aws_access_key_id": aws_access_key_id,
            "aws_secret_access_key": aws_secret_access_key,
        }

    def get_template_name(self):
        return 'aws.jinja2'

    def acquire_resources(self):
        region = self.get_free_region()

        if region is None:
            logging.error("No free region amongst selected ones: %s",
                          ', '.join(self.list_of_regions))
            raise Exception("No free region found")
        logging.info("Selected region %s", region)
        self.cluster_region = region

    def get_api_ip(self) -> str | None:
        return None

    def get_apps_ip(self):
        return None

    def post_installation(self):
        pass

    def get_free_region(self) -> str | None:
        """Finds first free region in provided list,
        if provided list is empty, it searches all regions"""
        candidates = self.list_of_regions[:]
        if len(candidates) == 0:
            candidates = boto3.Session().get_available_regions('ec2')
        for candidate in candidates:
            region = _get_connection(candidate, **self.boto_kwargs)
            try:
                count = len(region.describe_vpcs()['Vpcs'])
                if count < 5:
                    logging.debug("Selected region %s", candidate)
                    return candidate
            except ClientError:
                logging.debug("Skipping %s region, it might not be enalbed", candidate)
                continue
        return None



"""
Module implements merging of default configuration obtained via Dynaconf
with command line arguments.
"""
import argparse
import configparser
import logging
import warnings

from dynaconf import Dynaconf  # type: ignore[import-untyped]

ARCH_AMD = "amd64"
ARCH_X86_64 = "x86_64"
ARCH_ARM = "arm64"
ARCH_AARCH64 = "aarch64"
ARCH_PPC = "ppc64le"
ARCH_S390X = "s390x"

settings = Dynaconf(
    environments=True,
    lowercase_read=False,
    load_dotenv=True,
    settings_files=[name + end for name in ["settings", ".secrets"] for end in [".yaml", ".yml"]]
)


def _resolve_cloud_name(args: argparse.Namespace) -> dict | None:
    defaults = settings.as_dict()

    if defaults['CLOUD'][args.cloud].get('environments', None) is None:
        warnings.warn('[DEPRECATION WARNING] The structure of settings.yaml is changed, '
                      'please use environments list and cloud_env option. This behavior will be '
                      'removed in future releases.')
        return defaults['CLOUD'][args.cloud]
    default_env = defaults['CLOUD'][args.cloud].get('cloud_env', None) \
        if args.cloud_env is None else \
        args.cloud_env
    if default_env is None:
        logging.error("Couldn't resolve default environment")
        raise Exception("Invalid environment setup, cloud_env is missing")
    for env in defaults['CLOUD'][args.cloud]['environments']:
        if env['name'] == default_env:
            return env
    logging.warning("No environment found, maybe all variables are passed from command line")
    return None


def read_config(args: argparse.Namespace, default_args: dict) -> dict:
    """
    Reads config from Dynaconf and merges it with arguments provided via commandline.
    """
    result = {'cloud': None,
              'dns': None,
              'cloud_name': None,
              'cluster_name': args.cluster_name}
    if not args.__contains__('cloud'):
        return result
    defaults = settings.as_dict()

    if args.dns_provider is not None:
        result['dns'] = {'provider': args.dns_provider,
                         'conf': defaults['DNS'][args.dns_provider]}
        result['dns']['conf'].update(
            {j[4:]: i['proc'](vars(args)[j])
             for j, i in default_args['dns'].items()
             if vars(args)[j] is not None}
        )
        if 'credentials_file' in result['dns']['conf'].keys():
            config = configparser.ConfigParser()
            config.read(result['dns']['conf']['credentials_file'])

            result['dns']['conf'].update({
                "aws_access_key_id": config["default"]["aws_access_key_id"],
                "aws_secret_access_key": config["default"]["aws_secret_access_key"],
            })

    if args.cloud is not None:
        cloud_defaults = _resolve_cloud_name(args)
        result['cloud_name'] = args.cloud
        result['cloud'] = cloud_defaults
        result['cloud'].update(
            {j: i['proc'](vars(args)[j]) for j, i in default_args['install'].items()
             if vars(args)[j] is not None}
        )

        if 'credentials_file' in result['cloud'].keys():
            config = configparser.ConfigParser()
            config.read(result['cloud']['credentials_file'])

            result['cloud'].update({
                "aws_access_key_id": config["default"]["aws_access_key_id"],
                "aws_secret_access_key": config["default"]["aws_secret_access_key"],
            })

        if result['dns'] is not None:
            result['dns']['conf'].update({
                'cluster_name': args.cluster_name,
                'base_domain': result['cloud']['base_domain']
            })
    return result



#
# Copyright 2020 Osia authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The module represents all exported classes and functions
required to full installation of cluster"""
from .clouds import InstallerProvider
from .dns import DNSProvider
from .downloader import download_installer
from .executor import delete_cluster, install_cluster

__all__ = ['InstallerProvider',
           'DNSProvider',
           'install_cluster',
           'delete_cluster',
           'download_installer']



#
# Copyright 2020 Osia authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module contains basics and common functionality to set up DNS."""
from __future__ import annotations

import json
from abc import ABC, abstractmethod
from os import path
from pathlib import Path
from typing import Protocol

from osia.installer.clouds.base import _AbstractInstaller


class DNSProvider:
    """Class implements dynamic provier of DNSUtil base class"""

    __instance: DNSProvider | None = None

    # pylint: disable=protected-access
    @classmethod
    def register_provider(cls, name: str, clazz: type[_DNSUtil]):
        """Method to dynamically register new implementation of DNSUtil"""
        cls.instance().__add_provider(name, clazz)

    @classmethod
    def instance(cls) -> DNSProvider:
        """Returns singleton instance"""
        if cls.__instance is None:
            cls.__instance = cls()
        return cls.__instance

    def __init__(self) -> None:
        self.providers: dict[str, type[_DNSUtil]] = {}

    def __add_provider(self, name: str, clazz: type[_DNSUtil]):
        self.providers[name] = clazz

    def __getitem__(self, name: str) -> type[_DNSUtil]:
        return self.providers[name]

    def load(self, directory: str) -> _DNSUtil | None:
        """Method loads saved configuration of specific DNSUtil from
        file"""
        for k in self.providers:
            if path.exists(f"{directory}/{k}.json"):
                res = self[k]()
                res.unmarshall(directory)
                return res
        return None


class _DNSUtil(Protocol):
    """Protocol for DNSUtils"""

    @abstractmethod
    def add_api_domain(self, instance: _AbstractInstaller):
        """Method registers api domain in selected provider"""

    @abstractmethod
    def add_apps_domain(self, instance: _AbstractInstaller):
        """Method registers apps domain in selected provider"""

    @abstractmethod
    def delete_domains(self):
        """Method deletes all registered domains in provider"""

    @abstractmethod
    def provider_name(self):
        """Method stores current configuration on DNS provider to $provider_name.json"""

    def marshall(self, out_dir: str):
        """Method stores current configuration on DNS provider to $provider_name.json"""

    def unmarshall(self, in_dir: str):
        """Method loads stored configuration of DNS into provider object"""

    def delete_file(self):
        """Method removes unneeded configuration file"""


class DNSUtil(ABC):
    """Class implements basic settings for"""

    def __init__(self, cluster_name=None, base_domain=None, ttl=None):
        self.cluster_name = cluster_name
        self.base_domain = base_domain
        self.ttl = ttl
        self.modified = False

    @abstractmethod
    def add_api_domain(self, instance: _AbstractInstaller):
        """Method registers api domain in selected provider"""

    @abstractmethod
    def add_apps_domain(self, instance: _AbstractInstaller):
        """Method registers apps domain in selected provider"""

    @abstractmethod
    def delete_domains(self):
        """Method deletes all registered domains in provider"""

    @abstractmethod
    def provider_name(self):
        """Get name of provider"""

    def marshall(self, out_dir: str):
        """Method stores current configuration on DNS provider to $provider_name.json"""
        if not self.modified:
            return
        with open(f"{out_dir}/{self.provider_name()}.json", "w") as output:
            json.dump(self, output, default=lambda o: o.__dict__)

    def unmarshall(self, in_dir: str):
        """Method loads stored configuration of DNS into provider object"""
        with open(f"{in_dir}/{self.provider_name()}.json") as in_stream:
            dns_conf = json.load(in_stream)
            for k in dns_conf:
                setattr(self, k, dns_conf[k])

    def delete_file(self):
        """Method removes unneeded configuration file"""
        file_path = Path(self.cluster_name) / f"{self.provider_name()}.json"
        if file_path.exists():
            file_path.unlink()



#
# Copyright 2020 Osia authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module responsible for download of openshift-install binary"""
import logging
import platform
import re
import stat
import tarfile
import time
from pathlib import Path
from shutil import copyfileobj
from tempfile import _TemporaryFileWrapper

import requests
from bs4 import BeautifulSoup

from .utils import get_data

PROD_ROOT = "http://mirror.openshift.com/pub/openshift-v4/{}/clients/ocp/"
BUILD_ROOT = "https://openshift-release-artifacts.svc.ci.openshift.org/"
PREVIEW_ROOT = "http://mirror.openshift.com/pub/openshift-v4/{}/clients/ocp-dev-preview/"

VERSION_RE = re.compile(r"^openshift-install(-rhel(?P<rhel>\d+))?(-(?P<platform>(linux|mac)))?"
                        r"(-(?P<architecture>\w+))?(-(?P<version>\d+.*))?\.tar\.gz")
EXTRACTION_RE = re.compile(r'.*Extracting tools for .*, may take up to a minute.*')


def _current_platform() -> tuple[str, str]:
    if platform.system() == "Linux" and platform.machine() == "x86_64":
        return "linux", "amd64"
    if platform.system() == "Linux" and (
            platform.machine() == "arm64" or platform.machine() == "aarch64"):
        return "linux", "arm64"
    if platform.system() == "Darwin" and platform.machine() == "arm64":
        return "mac", "arm64"
    if platform.system() == "Darwin" and platform.machine() == "x86_64":
        return "mac", "amd64"

    raise Exception(f"Unrecognized platform {platform.system()} {platform.machine()}")


def get_url(directory: str, arch: str, fips: bool = False,
            rhel_version: str | None = None) -> tuple[str, str | None]:
    """Searches the http directory and returns both url to installer
    and version.
    """
    if fips and not rhel_version:
        raise Exception("Rhel version was not detected. Please download installer separatly.")

    logging.debug('Url for installers look-up %s', directory)
    lst = requests.get(directory, allow_redirects=True)
    tree = BeautifulSoup(lst.content, 'html.parser')
    links = tree.find_all('a')
    installer, version = None, None
    os_name, local_arch = _current_platform()
    for k in links:
        logging.debug('Parsing link: %s', k.get('href'))
        match = VERSION_RE.match(k.get('href'))

        if match:
            if match.group("version"):
                version = match.group('version')

            if fips and match.group("rhel") == rhel_version:
                installer = lst.url + k.get('href')
                break

            if not fips and match.group('platform') == os_name:
                if (local_arch == match.group('architecture')) \
                        or (local_arch == arch and not match.group('architecture')):
                    installer = lst.url + k.get('href')
                    break
    else:
        if fips:
            raise Exception(f"FIPS Installer not found for {rhel_version=}")
        raise Exception("Installer not found")
    return installer, version


def get_devel_url(version: str, arch: str, fips: bool = False,
                  rhel_version: str | None = None) -> tuple[str, str | None]:
    """
    Searches developement sources and returns url to installer
    """
    req = requests.get(BUILD_ROOT + version + "/", allow_redirects=True)
    ast = BeautifulSoup(req.content, 'html.parser')
    logging.info('Checking stage repository for installer')
    while len(ast.find_all('p')) != 0 and EXTRACTION_RE.match(next(ast.find_all('p')[0].children)):
        logging.debug('The installer was not extracted yet, waiting for 10s')
        time.sleep(10)
        req = requests.get(BUILD_ROOT + version, allow_redirects=True)
        ast = BeautifulSoup(req.content, 'html.parser')
    logging.debug('Installer found on page, continuing')
    return get_url(req.url, arch, fips, rhel_version)


def get_prev_url(version: str, arch: str, fips: bool = False,
                 rhel_version: str | None = None) -> tuple[str, str | None]:
    """Returns installer url from dev-preview sources"""
    return get_url(PREVIEW_ROOT.format(arch) + version + "/", arch, fips, rhel_version)


def get_prod_url(version: str, arch: str, fips: bool = False,
                 rhel_version: str | None = None) -> tuple[str, str | None]:
    """Returns installer url from production sources"""
    return get_url(PROD_ROOT.format(arch) + version + "/", arch, fips, rhel_version)


def _get_storage_path(version: str, install_base: str) -> str:
    path = Path(install_base)
    spec_path = path.joinpath(version)
    if not spec_path.exists():
        spec_path.mkdir()
    return spec_path.as_posix()


def _extract_tar(buffer: _TemporaryFileWrapper[bytes], target: str) -> Path:
    result = None
    with tarfile.open(buffer.name) as tar:
        inst_info = None
        for i in tar.getmembers():
            if i.name in ['openshift-install', 'openshift-install-fips']:
                inst_info = i
        if inst_info is None:
            raise Exception("error")
        stream = tar.extractfile(inst_info)
        if stream is None:
            raise OSError("could not extract tar")
        result = Path(target).joinpath(inst_info.name)
        with result.open('wb') as output:
            copyfileobj(stream, output)
        result.chmod(result.stat().st_mode | stat.S_IXUSR)
    return result


def get_installer(tar_url: str, target: str) -> str:
    """Download and extract the installer into the target"""
    return get_data(tar_url, target, _extract_tar)


# pylint: disable=too-many-arguments
def download_installer(installer_version: str,
                       installer_arch: str,
                       dest_directory: str,
                       source: str,
                       fips: bool = False,
                       rhel_version: str | None = None) -> str:
    """Starts search and extraction of installer"""
    logging.debug("Getting version %s of %s, storing to directory %s and devel is %r",
                  installer_version, installer_arch, dest_directory, source)

    downloader = None
    if source == "prod":
        downloader = get_prod_url
    elif source == "prev":
        downloader = get_prev_url
    elif source == "devel":
        downloader = get_devel_url
    else:
        raise Exception("Error for source profile " + source)

    url, version = downloader(installer_version, installer_arch, fips, rhel_version)
    logging.debug('Installer\'s URL is  %s and full version is %s', url, version)
    root = Path(dest_directory)
    if version:
        root = root.joinpath(version)

    installer_exe_name = 'openshift-install'

    if fips:
        installer_exe_name = 'openshift-install-fips'

    if root.exists() and root.joinpath(installer_exe_name).exists():
        logging.info('Found installer at %s', root.as_posix())
        return root.joinpath(installer_exe_name).as_posix()
    root.mkdir(parents=True, exist_ok=True)
    return get_installer(url, root.as_posix())



#
# Copyright 2020 Osia authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module implements dns methods to work with route53 provider"""
import logging

import boto3

from osia.installer.clouds.base import _AbstractInstaller
from osia.installer.dns.base import DNSUtil


def _get_connection(**kwargs):
    return boto3.client('route53', **kwargs)


class Route53Provider(DNSUtil):
    """Class implements DNSUtil base specific for route53"""
    def __init__(self, api_ip=None, apps_ip=None, aws_access_key_id=None, aws_secret_access_key=None, **kwargs):
        super().__init__(**kwargs)

        self.zone_id = None
        self.api_ip = api_ip
        self.apps_ip = apps_ip

        self.boto_kwargs = {
            "aws_access_key_id": aws_access_key_id,
            "aws_secret_access_key": aws_secret_access_key,
        }

    def provider_name(self):
        return 'route53'

    def _get_hosted_zone(self):
        if self.zone_id is None:
            zones = _get_connection(**self.boto_kwargs).list_hosted_zones()['HostedZones']
            result = [v['Id'] for v in zones if v['Name'] == (self.base_domain + ".")]
            if len(result) == 0:
                raise Exception(f"Unable to find hosted_zone {self.base_domain} in zone list.")
            self.zone_id = result[0]
        return self.zone_id

    def _execute_command(self, prefix: str, mode: str, ip_addr: str):
        change_batch = {
            'Changes': [
                {'Action': mode,
                 'ResourceRecordSet': {
                     'Name': '.'.join([prefix, self.cluster_name, self.base_domain]) + '.',
                     'Type': 'A',
                     'TTL': self.ttl,
                     'ResourceRecords': [
                         {'Value': ip_addr}
                     ]
                 }
                 }
            ]
        }
        conn = _get_connection(**self.boto_kwargs)
        try:
            conn.change_resource_record_sets(
                HostedZoneId=self._get_hosted_zone(),
                ChangeBatch=change_batch)
        except conn.exceptions.InvalidChangeBatch as ex:
            logging.warning("Exception thrown while %s operation, next steps will possibly fail",
                            mode.lower())
            logging.debug(ex)
        self.modified = True

    def add_api_domain(self, instance: _AbstractInstaller):
        self.api_ip = instance.get_api_ip()
        self._execute_command('api', 'CREATE', self.api_ip)

    def add_apps_domain(self, instance: _AbstractInstaller):
        self.apps_ip = instance.get_apps_ip()
        self._execute_command('*.apps', 'CREATE', self.apps_ip)

    def delete_domains(self):
        if self.api_ip is not None:
            self._execute_command('api', 'DELETE', self.api_ip)
        if self.apps_ip is not None:
            self._execute_command('*.apps', 'DELETE', self.apps_ip)
        self.delete_file()



#
# Copyright 2020 Osia authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module implements command line interface for the installation of
openshift"""
import argparse
import logging
from subprocess import Popen

import coloredlogs  # type: ignore[import-untyped]
import distro
from semantic_version import (SimpleSpec,  # type: ignore[import-untyped]
                              Version)

from .config import read_config
from .config.config import (ARCH_AARCH64, ARCH_AMD, ARCH_ARM, ARCH_PPC,
                            ARCH_S390X, ARCH_X86_64)
from .installer import (delete_cluster, download_installer, install_cluster,
                        storage)


def _identity(in_attr: str) -> str:
    return in_attr


def _read_list(in_str: str) -> list[str]:
    return in_str.split(',')


ARGUMENTS: dict = {
    'common': {
        'cloud': {'help': 'Cloud provider to be used.', 'type': str,
                  'choices': ['openstack', 'aws']},
        'cloud_env': {'help': 'Environment of cloud to be used.', 'type': str},
        'dns_provider': {'help': 'Provider of dns used with openstack cloud',
                         'type': str, 'choices': ['nsupdate', 'route53']},
        'credentials_file': {'help': 'AWS config file for ec2, using "default" profile'},
    },
    'install': {
        'os_image': {'help': 'Image to override', 'type': str},
        'base_domain': {'help': 'Base domain for the cluster', 'type': str},
        'master_flavor': {'help': 'Flavor used for master nodes'},
        'master_replicas': {'help': 'Number of replicas for master nodes', 'type': int},
        'pull_secret_file': {'help': 'File containing pull secret from `cloud.redhat.com`'},
        'ssh_key_file': {'help': 'File with ssh key used by installer'},
        'list_of_regions': {'help': 'List of regions, comma separated values', 'proc': _read_list},
        'osp_cloud': {'help': 'Name of openstack cloud identity in clouds.yaml'},
        'osp_base_flavor': {'help': 'Base flavor to be used in openstack'},
        'osp_image_download': {'help': 'Enable image download by osia instead of openshift-install',
                               'action': 'store_true'},
        'osp_image_unique': {'help': 'Upload unique image per cluster',
                             'action': 'store_true'},
        'network_list': {'help': 'List of usable openstack networks, comma separated values',
                         'proc': _read_list},
        'worker_flavor': {'help': 'flavor of worker node'},
        'worker_replicas': {'help': 'Number of replicas of worker nodes', 'type': int},
        'certificate_bundle_file': {'help': 'CA bundle file'},
        'images_dir': {'help': 'Directory where images should be stored', 'type': str,
                       'default': 'images'},
        'skip_clean': {'help': 'Skip clean when installation fails', 'action': 'store_true'},
        'enable_fips': {'help': 'Enable fips mode to the cluster', 'action': 'store_true'},
    },
    'deprecated': {
        'psi_cloud': {'help': 'DEPRECATED see osp_cloud'},
        'psi_base_flavor': {'help': 'DEPRECATED see osp_base_flavor'},
    },
    'dns': {
        'dns_ttl': {'help': 'TTL of the records', 'type': int},
        'dns_key_file': {'help': 'Keyfile used to access dns server via nsupdate'},
        'dns_credentials_file': {'help': 'AWS config file for route53, using "default" profile'},
        'dns_zone': {'help': 'Zone on server where the record will be stored'},
        'dns_server': {'help': 'Address of server with running bind'},
        'dns_use_ipv4': {'help': 'Use only IPv4 for DNS settings', 'action': 'store_const',
                         'const': True},
    }
}

for a in [v for _, x in ARGUMENTS.items() for u, v in x.items()]:
    if not a.get('proc', None):
        a['proc'] = _identity


def _check_fips_compatible(rhel_version: str | None) -> tuple[bool, str | None]:
    if not rhel_version:
        return False, "FIPS installs are supported only from RHEL systems"
    try:
        with Popen(["fips-mode-setup", "--check"]) as proc:
            proc.wait()
            if proc.returncode != 0:
                return False, "FIPS is not enabled on the system"
    except FileNotFoundError:
        return False, "fips-mode-setup must be installed on the system"

    return True, None


def _resolve_installer(from_args):
    if from_args.installer is None and from_args.installer_version is None:
        raise Exception('Either installer or installer-version must be passed')
    if from_args.installer:
        return from_args.installer

    rhel_version = None
    if distro.id() == "rhel":
        rhel_version = distro.major_version()

    if from_args.enable_fips:
        supported, msg = _check_fips_compatible(rhel_version)
        if not supported:
            raise Exception(msg)

        if Version.coerce(from_args.installer_version.split("-")[-1]) in SimpleSpec("<4.16"):
            # ocp <4.16 does not have dedicated openshift-install-fips, it is
            # fine to run normal installer on FIPS enabled RHEL
            from_args.enable_fips = False

    return download_installer(from_args.installer_version,
                              from_args.installer_arch,
                              from_args.installers_dir,
                              from_args.installer_source,
                              rhel_version=rhel_version,
                              fips=from_args.enable_fips)


def _merge_dictionaries(from_args):
    result = read_config(from_args, ARGUMENTS)
    result["installer"] = _resolve_installer(from_args)
    if result.get('cloud'):
        # pylint: disable=unsupported-assignment-operation
        result['cloud']['installer'] = result['installer']
    return result


def _exec_install_cluster(args):
    conf = _merge_dictionaries(args)
    if not args.skip_git:
        storage.check_repository()
    logging.info('Starting the installer with cloud name %s', conf['cloud_name'])
    install_cluster(
        conf['cloud_name'],
        conf['cluster_name'],
        conf['cloud'],
        conf['installer'],
        dns_settings=conf['dns']
    )
    if not args.skip_git:
        storage.write_changes(conf['cluster_name'])


def _exec_delete_cluster(args):
    # cleanup of fips cluster can be done from anywhere
    args.enable_fips = None

    conf = _merge_dictionaries(args)

    if not args.skip_git:
        storage.check_repository()

    delete_cluster(conf['cluster_name'], conf['installer'])

    if not args.skip_git:
        storage.delete_directory(conf['cluster_name'])


def _get_helper(parser: argparse.ArgumentParser):
    def printer(unused_conf):
        print("Operation not set, please specify either install or clean!")
        parser.print_help()

    return printer


def _create_commons() -> argparse.ArgumentParser:
    commons = argparse.ArgumentParser(add_help=False)
    common_arguments: list[tuple[list[str], dict]] = [
        (['--cluster-name'], dict(required=True, help='Name of the cluster')),
        (['--installer'], dict(required=False,
                               help='Executable binary of openshift install cli', default=None)),
        (['--installer-version'], dict(help='Version of downloader to be downloaded',
                                       default='latest', type=str)),
        (['--installer-arch'], dict(help='Architecture of downloader to be downloaded',
                                    choices=[ARCH_AMD, ARCH_X86_64, ARCH_ARM,
                                             ARCH_AARCH64, ARCH_PPC, ARCH_S390X],
                                    default=ARCH_AMD, type=str)),
        (['--installer-source'], dict(type=str,
                                      help='Set the source to search for installer',
                                      choices=["prod", "devel", "prev"],
                                      default='prod')),
        (['--installers-dir'], dict(help='Folder where installers are stored',
                                    required=False, default='installers')),
        (['--skip-git'], dict(help='When set, the persistance will be skipped',
                              action='store_true')),
        (['-v', '--verbose'], dict(help='Increase verbosity level', action='store_true')),
    ]
    for args, kwargs in common_arguments:
        commons.add_argument(*args, **kwargs)
    return commons


def _setup_parser() -> argparse.ArgumentParser:
    commons = _create_commons()

    parser = argparse.ArgumentParser("osia")
    parser.set_defaults(func=_get_helper(parser))
    sub_parsers = parser.add_subparsers()

    install = sub_parsers.add_parser('install', help='Install new cluster', parents=[commons])

    for arg, value in sorted({k: v for _, x in ARGUMENTS.items() for k, v in x.items()}.items()):
        install.add_argument(f"--{arg.replace('_', '-')}",
                             **{k: v for k, v in value.items() if k != 'proc'})
    install.set_defaults(func=_exec_install_cluster)

    clean = sub_parsers.add_parser('clean', help='Remove cluster', parents=[commons])
    clean.set_defaults(func=_exec_delete_cluster)
    return parser


def main_cli() -> None:
    """Function represents main entrypoint for the
    osia installer

    It sets up the cli and starts the executor."""
    parser = _setup_parser()
    args = parser.parse_args()

    coloredlogs.DEFAULT_FIELD_STYLES.update(filename=dict(color='blue'))
    fmt: str = "%(asctime)s %(filename)12s:%(lineno)-5i %(levelname)-8s %(message)s"

    if vars(args).get('verbose', False):
        coloredlogs.install(fmt=fmt, level=logging.DEBUG)
    else:
        coloredlogs.install(fmt=fmt, level=logging.INFO)

    logging.captureWarnings(True)

    for package in ["urllib3", "git"]:
        logging.getLogger(package).setLevel(logging.INFO)

    args.func(args)
